package minicp.enigma;


public class EnigmaEmulator {

    private Alphabet alphabet;
    private Rotor[] rotors;
    private int[] plug_board;
    private int[] reflector;

    public EnigmaEmulator(Alphabet alphabet, int[][] perms, int[] plug_board, int[] reflector) {
        this.alphabet = alphabet;
        rotors = new Rotor[perms.length];
        for (int i = 0; i < perms.length; ++i) {
            assert perms[i].length == alphabet.size();
            rotors[i] = new Rotor(perms[i], 0);
        }
        this.plug_board = new int[alphabet.size()];
        this.reflector = new int[alphabet.size()];
        for (int i = 0; i < alphabet.size(); ++i) {
            this.plug_board[i] = plug_board[i];
            this.reflector[i] = reflector[i];
        }
    }

    public Alphabet get_alphabet() {
        return alphabet;
    }

    public int num_rotors() {
        return rotors.length;
    }

    public int[][] get_forward_rotors() {
        int[][] values = new int[rotors.length][];
        for (int i = 0; i < rotors.length; ++i) {
            values[i] = rotors[i].get_forward();
        }
        return values;
    }

    public int[][] get_backward_rotors() {
        int[][] values = new int[rotors.length][];
        for (int i = 0; i < rotors.length; ++i) {
            values[i] = rotors[i].get_backward();
        }
        return values;
    }

    public int[] get_reflector() {
        int[] r = new int[reflector.length];
        System.arraycopy(reflector, 0, r, 0, r.length);
        return r;
    }

    public void set_key(int[] key) {
        assert key.length == rotors.length;
        for (int i = 0; i < key.length; ++i) {
            rotors[i].set_key(key[i]);
        }
    }

    public int get_key(int i) {
        return rotors[i].get_position();
    }

    public void set_plugs(int[] plugs) {
        assert plug_board.length == plugs.length;
        for (int i = 0; i < plug_board.length; ++i) {
            plug_board[i] = plugs[i];
        }
    }

    public void step() {
        for (int i = 0; i < rotors.length && rotors[i].step(); ++i) ;
    }

    public int encryption(int c) {
        int x = plug_board[c];

        for (Rotor r : rotors)
            x = r.encrypt_forward(x);

        x = reflector[x];

        for (int i = rotors.length - 1; i >= 0; --i)
            x = rotors[i].encrypt_backward(x);

        return plug_board[x];
    }

    public char encrypt_char(char c) {
        int i = encryption(alphabet.get_digit(c));
        step();
        return alphabet.get_char(i);
    }

    public String encrypt(String plaintext) {
        String ciphertext = "";
        char[] chars = plaintext.toCharArray();
        for (char c : chars) {
            ciphertext += encrypt_char(c);
        }
        return ciphertext;
    }
/*
    public static void main(String[] args) {
        Alphabet alphabet = new Alphabet("ABCDEFGHIJKLMNOPQRSTUVWXYZ");
        int[] r1 = Permutation.random(alphabet.size());
        int[] r2 = Permutation.random(alphabet.size());
        int[] r3 = Permutation.random(alphabet.size());

        int[] rf = Permutation.random_symmetric(alphabet.size() / 2, alphabet.size());
        int[] pb = Permutation.random_symmetric((alphabet.size() - 6) / 2, alphabet.size());


        EnigmaEmulator machine = new EnigmaEmulator(alphabet, new int[][]{r1, r2, r3}, pb, rf);


        String plaintext = "TESTOFMYJAVAENGIMAEMULATORISTHISGOINGTOSUCCEEDORAMIGOINGTODEBUGITFORBLOODYAGES";

        machine.set_key(new int[]{4, 10, 13});
        String ciphertext = machine.encrypt(plaintext);

        machine.set_key(new int[]{4, 10, 13});
        String replaintext = machine.encrypt(ciphertext);

        assert plaintext == ciphertext;

        System.out.println(plaintext);
        System.out.println(ciphertext);
        System.out.println(replaintext);
    }
 */
}

